=head1 NAME
 
Parser - Parses mammography text field extracts BI-RADS info.

Published in:
Information extraction for clinical data mining: A mammography case study
Nassif, H., Wood, R., Burnside, E. S., Ayvaci, M., Shavlik, J., and Page, D.
9th IEEE International Conference on Data Mining (ICDM'09) Workshops, Miami, pp. 37-42, 2009

Extended for breast composition in:
Automatic Classification of Mammography Reports by BI-RADS Breast Tissue Composition Class
Percha B, Nassif H, Lipson J, Burnside E, and Rubin D
Journal of the American Medical Informatics Association (JAMIA), 19(5):913-916, 2012.

=cut

=head1 VERSION

Version 12

Uses BiRads version 3 (Breast Imaging Reporting and Data System (BI-RADS), American College of Radiology, Reston, VA, 1998)

The feature titles follows our group convention, in /u/mammobayes/data/BIRADSv4

Mass size: invokes MassSize.pm
Returns the measurement set with the highest number of dimensions, up to three dimensions. If multiple, returns the largest one.
E.g.: "abnormality of 3 mm x 4 cm is encountered in left breast. Abnormalities of 4x2x1 cm and 2x5x1 cm in right breast"
returns: Mass Size 1 = 50 mm, Mass Size 2 = 20 mm, Mass Size 3 = 10 mm
Includes breast composition & palpability

=cut

package Parser;

use strict;
use warnings;
use Carp;
use MassSize;

=head2 new()

 Title   : new
 Usage   : my $parser = Parser->new('-inputFile' => $textFile,
			            '-outputFile' => $outputFile,
                                    '-fieldDelimiter' => "\t",
                                    '-reportNB' => 21,
                                    '-binary' => 0,
			           );
 Function: Returns a Parser object.
 Returns : A new Parser object, whose structrure is the following:
           $self->{'INPUT FILE'}; # the original ucsf data
                ->{'OUTPUT FILE'}; # the output file name
                ->{'FIELD DELIMITER'} = "\t"; # to convert from to excell
                ->{'FEATURE FIELDS'}; #new feature fields added
 Args    : -inputFile: input file name
           -outputFile: output file name
           -fieldDelimiter: the field delimiter.
           -reportNB: the column number of the report field.
           -binary: if the new field output is binary or count.

=cut

sub new{
    my $class = shift;
    my %arguments = @_;


    my $self = {}; # a reference to hash
    bless $self, $class;

    # store the arguments
    $self->{"INPUT FILE"} = $arguments{'-inputFile'} ||  
	confess "should use -inputFile. stopped";
    $self->{"OUTPUT FILE"} = $arguments{'-outputFile'} ||  
	confess "should use -outputFile. stopped";
    $self->{"FIELD DELIMITER"} = $arguments{'-fieldDelimiter'} ||
        confess "should use -fieldDelimiter. stopped";
    $self->{"REPORT_NB"} = $arguments{'-reportNB'};
    confess "should use -reportNB. stopped" if !defined $self->{"REPORT_NB"};
    $self->{"BINARY"} = $arguments{'-binary'};
    confess "should use -binary. stopped" if !defined $self->{"BINARY"};


    # HoA storing the feature fields.
    my $space = '[\s,\-\(\):\\\/\']';
    my $word = '((\w|\d(\.\d)?)*?'.$space.'+?)';
    $self->{"WORD"} = $word;
    my %HoA = (
	'1.shape' => [ '\bround(?!ing)'.$word.'{0,5}?(dens|mass|struc|asym|nodule)', 
		       '\b(oval|ovoid)\b'.$word.'{0,10}?(dens|mass|struc|asym|nodule)|(mass|nodule)'.$word.'{0,5}?\b(oval|ovoid)\b',
		       '\b(macro|multi|bi)?lobula\w*\s*(?!'.$word.'*(carcinoma|cancer|plasia))',
		       '\birregular(?!'.$word.'{0,3}?\w*calcif\w*ion)'],
	'2.margins' => ['circumscribed|\bwell.{0,2}?(defined|marginated)\b|\bsharply\b|\bsharp'.$word.'{0,3}?margin', 
		      '\bmicrolobulat', 
		      '\bobscured\b'.$word.'*?(mass|margin)|(mass|margin)'.$word.'*?\bobscured\b', 
		      'indistinct|\bill.{0,2}?defined\b', 
		      'spiculat'],
	'3.density' => ['\bhigh'.$word.'{0,3}?density\b', 
		      'isodense|\bequal.{0,2}?density\b', 
		      '\blow'.$word.'{0,2}?density\b', 
		      'fat.{0,2}?contain|radio.{0,2}?luce'],
	'4.morphology' => ['\bskin.{0,2}?(?!non)\w*calcif', 
			 'vascular.{0,2}?(?!non)\w*calcif', 
			 'coarse|popcorn', 
			 'rod.{0,2}?like', 
			 '\bround(?!ing)'.$word.'{0,5}?(?!non)\w*calcif(?!ied)|(?<!non)calcif(?!ied)'.$word.'{0,5}?\bround(?!ing)', 
			 '\blucen'.$word.'*?(?!non)\w*calcif|(?<!non)calcif'.$word.'*?\blucen', 
			 '\begg.{0,2}?shell\b|\brim', 
			 '\bmilk\b', 
			 '\bsuture\b', 
			 '\bdystrophic\b', 
			 'punctate', 
			 'amorph', 
			 'pl\w*mor|heterogen'.$word.'{0,3}?(?!non)\w*calcif', 
			 '\bfine'.$word.'{0,5}?(?!non)\w*calcif|\bcasting\b|branch'.$word.'{0,5}?(?!non)\w*calcif|(?<!non)calcif'.$word.'{0,5}?branch'],
	'5.distribution' => ['\bgroup'.$word.'*?(?!non)\w*calcif|(?<!non)calcif\w*'.$word.'*?\bgroup|\bcluster'.$word.'{0,5}?(?!non)\w*calcif|(?<!non)calcif'.$word.'{0,7}?\bcluster', 
			   '\blinear'.$word.'{0,5}?(?!non)\w*calcif|(?<!non)calcif'.$word.'{0,5}?\blinear', 
			   'segmental', 
			   'regional(?!'.$word.'{0,3}(medical|hospital|center|enhanc))', 
			   '(diffuse|scatter)(?!'.$word.'{0,2}(dens|fibro))'.$word.'*?(?!non)\w*calcif|(?<!non)calcif(?!'.$word.'{0,2}(dens|fibro))'.$word.'*?(diffuse|scatter)'],
	'6.special.cases' => ['\btubular\b(?!'.$word.'{0,3}carcinoma)|duct'.$word.'{0,5}?dilat\w*|dilat'.$word.'{0,5}?duct', 
			    'mamm?ar\w*.{0,2}?lymph\w*.{0,2}?node',
			    '(tis|asym)'.$word.'{0,2}?(tis|breast|asym)'.$word.'{0,2}?(tis|breast|asym)',
			    'focal'.$word.'{0,4}?asym|asym'.$word.'{0,4}?focal'],   
	'7.associated.findings' => ['skin'.$word.'{0,4}?retract|retract'.$word.'{0,4}?skin', 
				  'nipp?l'.$word.'{0,4}?retract|retract'.$word.'{0,4}?nipp?l',
				  'skin'.$word.'*?thick|thick'.$word.'*?skin',
				  'trabec'.$word.'*?thick|thick'.$word.'*?trabec',
				  'skin'.$word.'*?lesion|lesion'.$word.'*?skin',
				  'axillary'.$word.'{0,2}?adenopathy', 
				  'arch\w*.{0,2}?distor'
				 ],
	'8.breast.composition' => ['((predomin\w*|entire\w*|primar\w*|mostly|largely|completely|relative\w*)'.$space.$word.'{0,2}?fat(ty)?)|(breasts?\s+'.$word.'?\b(is|are)\b\s+fat(ty)?\b)(?!'.$word.'*fibro.?(gland|nodul))',
				   'scatter\w*'.$space.$word.'{0,3}?fibro.?(gland|nodul)',
				   '((mildly|moderat\w*|heterogen\w*)'.$space.$word.'{0,2}?(dens|breast|tissue|parenchyma))|(focall?y\s+dense)',
				   '((extrem\w*|homogen\w*ly|\bvery\b)\s+dens)|(breasts?\s+'.$word.'?\b(is|are)\b\s+dense\b)|(significantly dense\b)',
				   ],
	'9.palpability' => ['(?<!non)palp'],       
	);
    $self->{"FEATURE FIELDS"} = \%HoA;

    my %titles = ( #the titles of the above fields
	'1.shape' => ["ShRo","ShOv","ShLob","ShIr"],
	'2.margins' => ["MgCrc","MgMl","MgObs","MgIll","MgSpi"],
	'3.density' => ["DenHigh","DenIso","DenLow","DenFat"],
	'4.morphology' => ["CalSkin","CalVas","CalPcrn","CalRod","CalRnd","CalLcn","CalRim",
			 "CalMilk","CalSut","CalDyst","CalPu","CalAmrp","CalPl","CalFiLi"],
	'5.distribution' => ["DistGrpd","DistLi","DistSg","DistRgn","DistDiff"],
	'6.special.cases' => ["SCTub","SCLmph","SCAsym","SCFoc"],   
	'7.associated.findings' => ["SkRtr","NpRtr","SkThck","TrThkc","SkLes","AxAdnp","ArcDst"],
	'8.breast.composition' => ["BrCompFat","BrCompSctr","BrCompHtr","BrCompExtr"],   
	'9.palpability' => ["Palpable"],
	);
    $self->{"FEATURE TITLES"} = \%titles;

    #for negation
    my $d = 5;
    my %degrees = ( #the degrees of the above fields, ie how many words bet negation and attributes
	'1.shape' => [$d, 9, $d, 5],
	'2.margins' => [5, $d, $d, $d, 4],
	'3.density' => [$d, $d, $d, $d],
	'4.morphology' => [4, 5, $d, $d, 5, $d, $d,
			   5, $d, $d, $d, 5, 5, $d],
	'5.distribution' => [8, $d, 4, $d, 3],
	'6.special.cases' => [5, 4,  $d,  $d], 
	'7.associated.findings' => [4, 4, 5, $d, 5, 4, 9],
	'8.breast.composition' => [1, 1, 1, 1],
	'9.palpability' => [4],
	);
    $self->{"DEGREES"} = \%degrees;

    return $self;
}


=head2 extract()

 Title   : extract
 Usage   : $parser->extract();
 Function: Extracts info from text and stores it in table columns (new fields)
 Returns : Writes down the output file, populates the new fields.
 Args    : -rearrange: rearranges the excell sheet and inserts the new fields.
                       default. Else outputs only the new fields.

=cut

sub extract{
    #store args
    my $self = shift;
    my %arguments = @_;
    my $rearrange = $arguments{"-rearrange"};
    $rearrange = 1 if !defined $rearrange;

    open(FILE, $self->{"INPUT FILE"}) ||
	confess "couldn't open ", $self->{"INPUT FILE"}, ": $!. Stopped"; 
    open(OUTPUT, '>', $self->{"OUTPUT FILE"}) ||
	confess "couldn't open ", $self->{"OUTPUT FILE"}, ": $!. Stopped";

    my $featureFieldsNb;
    #adds the nb of feature fields in each array of FEATURE FIELDS
    $featureFieldsNb += scalar @{$_} foreach values %{$self->{"FEATURE FIELDS"}};

    #handle title
    my $line = <FILE>; #skip first line
    my @title = $self->_handle_title($line, $rearrange);
    print OUTPUT join($self->{"FIELD DELIMITER"}, @title),"\n";

    my (@feature_fields, @massSize); #stores the extracted features
    my @totalFields;
    my (@divide, $report);
    while ($line = <FILE>){
	chomp $line;
	@divide = split (/$self->{"FIELD DELIMITER"}/, $line);
	$report = $divide[$self->{"REPORT_NB"}];
	@feature_fields = $self->_parse($report); 
	@massSize = @{&MassSize::mssz($report)};
	@totalFields = (@feature_fields, @massSize); #the new extracted features

	if ($rearrange){#rearrange fields
	    $divide[$_+@totalFields] =  $divide[$_] foreach ($self->{"REPORT_NB"}+1 .. $#divide);
	    $divide[$self->{"REPORT_NB"}+1+$_] = $totalFields[$_] foreach 0..$#totalFields;
	    print OUTPUT join($self->{"FIELD DELIMITER"}, @divide),"\n";
	}
	else{#just print new fields
	    print OUTPUT join($self->{"FIELD DELIMITER"}, @totalFields),"\n";
	}
    }

    close (FILE);
    close (OUTPUT);
}


=head2 _handle_title()

 Title   : _handle_title
 Usage   : $self->_handle_title($line);
 Function: Creates the title (first line) of the output file. 
           Incorporates the new fields amidst the old ones.
 Returns : An array of the title fields.
 Args    : the first line/row (table title)
           rearrange or not

=cut

sub _handle_title{
    my $self = shift;
    my $line = shift;
    my $rearrange = shift;

    chomp $line;
    my @divide = split (/$self->{"FIELD DELIMITER"}/, $line);

    my @totalFields;
    #store the features extracted from text
    foreach my $key (sort keys %{$self->{"FEATURE TITLES"}}){
	foreach my $field (@{$self->{"FEATURE TITLES"}{$key}}){
	    push @totalFields, "\"$field\"";
	}
    }
    
    #store the size
    push @totalFields, "\"mass size 1 (mm)\"", "\"mass size 2 (mm)\"", "\"mass size 3 (mm)\"";
    #push @totalFields, "\"MS1\"", "\"MS2\"", "\"MS3\"";

    if ($rearrange){#rearrange fields
	$divide[$_+@totalFields] =  $divide[$_] foreach ($self->{"REPORT_NB"}+1 .. $#divide);
	$divide[$self->{"REPORT_NB"}+1+$_] = $totalFields[$_] foreach 0..$#totalFields;
	return @divide;
    }
    else{
	return @totalFields;
    }
}


=head2 _parse()

 Title   : _parse
 Usage   : $self->_parse($report);
 Function: Parses the text and searches for the required keywords.
           Populates the new fields in integer mode.
 Returns : An integer array for the new feature fields, counting nb of occurences.
 Args    : the report (text from which to extract data)

=cut

sub _parse{
    my $self = shift;
    my $report = "\L$_[0]\E"; #convert to lower case letters
    my $word = $self->{"WORD"};

    my @array;
    my $i = 0;
    foreach my $key (sort keys %{$self->{"FEATURE FIELDS"}}){
	foreach my $j (0..$#{$self->{"FEATURE FIELDS"}{$key}}){
	    my $field = '\b\w*('.$self->{"FEATURE FIELDS"}{$key}[$j].')\w*\b';
	    {
		my $ultra = '(\w*echo\w*c|\w*sonog\w*)'; 
		my $negation = '\b(no|(?<!\bwhere )not|without)\b';
		my $doubleNegation = $negation.'(?='.$word.'{0,'.(${$self->{"DEGREES"}{$key}}[$j]).'}?(change\w*|all|correlat\w*|differ\w*|other|'.$ultra.'))';

#		my $searchForReject = '('.$negation.'|'.$ultra.')'.$word.'{0,(?(4)'.${$self->{"DEGREES"}{$key}}[$j].'|'.${$self->{"DEGREES"}{$key}}[$j].')}?('.$field.')';

		my $searchForReject = '(('.$negation.'|'.$ultra.')(?(5)'.$word.'{0,'.(${$self->{"DEGREES"}{$key}}[$j]-2).'}?|'.$word.'{0,'.${$self->{"DEGREES"}{$key}}[$j].'}?)('.$field.'))|('.$field.$word.'{0,5}'.$ultra.')';
		my $searchForAccept = $doubleNegation.$word.'{0,'.${$self->{"DEGREES"}{$key}}[$j].'}?('.$field.')';

		$array[$i] = 0;
		if ($self->{"BINARY"}) {
		    $array[$i] = 1 if $report =~ /$field/; # $field is present
		    $array[$i] = 0 if $report =~ /$searchForReject/; # $field is negated
		    $array[$i] = 1 if $report =~ /$searchForAccept/; # $field is double negated
		    while ($report =~ /($field)/g){
			my $core = $1;
			$array[$i] = 0 if ($core =~ /$negation|$ultra/); # negation within $field
			$array[$i] = 1 if ($core =~ /$doubleNegation/); # double negation within field
		    }
		}
		else { #not binary
		    while ($report =~ /($field)/g){
			my $core = $1;
			$array[$i]++; #field is present
			$array[$i]-- if ($core =~ /$negation|$ultra/); # negation within $field
			$array[$i]++ if ($core =~ /$doubleNegation/);
			#if ($core =~ /$negation|$ultra/){$array[$i]-- ; print "-: $core\n";}
			#if ($core =~ /$doubleNegation/){$array[$i]++ ; print "+: $core\n";}
		    }#TO DO: check if I go negative. Am i counting the same twice? 
		    while ($report =~ /($searchForReject)/g){ # $field is negated
			$array[$i]--;
			#print "-: $1\n";
		     }
		    while ($report =~ /($searchForAccept)/g){ # $field is double negated
			 $array[$i]++;
			 #print "+: $1\n";
		     }
		}
	    }#to stop $1 interference
	    $i++;
	}
    }
    return @array;
}


1;
